/* 
 * File:   Alignment.cpp
 * Author: tg
 * 
 * Created on November 6, 2011, 12:58 PM
 *
 * mrtailor-gui: Qt-based GUI for the program "mrtailor"
 * Copyright (C) 2012, Tim Gruene (tg@shelx.uni-ac.gwdg.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "AlignLayout.h"
#include "ReadClustalAln.h"
#include "guiException.h"
#include "PDB.h"

#include <iostream>
#include <string>

AlignLayout::AlignLayout(QWidget* parent)
: QWidget(parent){
    QLabel* label               = new QLabel (tr("Alignment File"));
    QPushButton* browseButton   = new QPushButton(tr ("Browse"));
    QHBoxLayout* toplayout      = new QHBoxLayout;
    QHBoxLayout* labellayout    = new QHBoxLayout;
    QHBoxLayout* alignmentlayout= new QHBoxLayout;

    alignmentInput_             = new QLineEdit (alignment_file_);
    vlayout_                    = new QVBoxLayout;
    targetIDSelector_           = new QComboBox;

    //aliViewer_                  = new AliViewer;
    aliViewer_                  = new AliTableView;

    targetIDSelector_->addItem(tr("Target ID (please load an alignment file)"));

    alignmentInput_->setStatusTip(tr("Filename with sequence alignment"));

    connect (alignmentInput_, SIGNAL(textEdited(const QString)),
            this, SLOT(nameChange(const QString)));
    connect (browseButton, SIGNAL (pressed()), this, SLOT (fileBrowsing()));
    connect (targetIDSelector_, SIGNAL(currentIndexChanged (const QString&)),
            this, SLOT(targetID(const QString&)));
    connect (this, SIGNAL(new_sequence(const std::map<std::string, std::string>&)),
            aliViewer_, SLOT(showAlignments(const std::map<std::string, std::string>&)));
    connect (this, SIGNAL (match(const std::string&, const std::string&)),
            aliViewer_, SLOT(showMatches(const std::string&, const std::string&)));

    toplayout->addWidget(label);
    toplayout->addWidget(alignmentInput_);
    toplayout->addWidget(browseButton);

    labellayout->addWidget(new QLabel(tr("Sequence Alignment")), 1, Qt::AlignCenter);

    alignmentlayout->addWidget(targetIDSelector_, 5, Qt::AlignTop);

    vlayout_->addLayout(toplayout, Qt::AlignCenter);
    vlayout_->addLayout(labellayout);
    vlayout_->addWidget(aliViewer_);
    vlayout_->addWidget(new QLabel(tr("ID of target sequence")), 1, Qt::AlignCenter);
    vlayout_->addLayout(alignmentlayout, Qt::AlignCenter);

    setLayout(vlayout_);
}

AlignLayout::~AlignLayout() {
}

void AlignLayout::nameChange(const QString& name) {
    alignment_file_ = name;
}

void AlignLayout::fileBrowsing() {
    ReadClustalAln* alignment (NULL);
    QString filename (QFileDialog::getOpenFileName(this, tr("Open Sequence Alignment"),
            alignment_file_, tr("Alignment file (*.aln)"), &alignFilter_));
    if (filename.isEmpty()) {
        return;
    }
    alignmentInput_->setText(filename);
    try {
        alignment = new ReadClustalAln (filename.toStdString().c_str());
        
    }
    catch (guiException::Format& f) {
        std::cerr << "Unable to read Clustal alignment from "
                << filename.toStdString()
                << "\n"
                << "More formats will be supported in a future version...\n";
        alignmentInput_->clear();
    }
    seqs_ = alignment->sequences();

    targetIDSelector_->clear();

    std::map<std::string, std::string>::const_iterator it;
    for (it=seqs_.begin(); it != seqs_.end(); ++it) {
        targetIDSelector_->addItem(tr(it->first.c_str()));
    }

    emit new_sequence(seqs_);
}

void AlignLayout::highlightMatches(std::string model_seq_ID, std::string pdb_chains) {
    model_sequence_ID_ = model_seq_ID;
    pdbChains_         = pdb_chains;
    
    if (pdbChains_.empty()) {
        return;
    }

    emit match(model_sequence_ID_, pdbChains_);
}

void AlignLayout::showalignments() {
}

/**
 * Signal forwarding from ComboBox selection
 * @param id
 */
void AlignLayout::targetID(const QString& id) {
    target_sequence_ID_ = id.toStdString();
    emit targetID(target_sequence_ID_);
}

/**
 * for a IDs in idschain select the one which is longest
 * @return 
 */
Sequence AlignLayout::modelSequence() const {
    std::map<std::string, std::string>::const_iterator it;
    it = seqs_.find(model_sequence_ID_);

    if (it == seqs_.end()) {
        // better: throw something or emit a signal: no match found, or similar
        return Sequence();
    }
    else {
        return Sequence(it->first, it->second);
    }
}

Sequence AlignLayout::targetSequence() const {
    std::map<std::string, std::string>::const_iterator it;
    it = seqs_.find(target_sequence_ID_);

    if (it == seqs_.end()) {
        // better: throw something or emit a signal: no match found, or similar
        return Sequence();
    }
    else {
        return Sequence(it->first, it->second);
    }
}
