/* 
 * File:   MRTailorGui.cpp
 * Author: tg
 * 
 * Created on November 7, 2011, 1:04 PM
 *
 * mrtailor-gui: Qt-based GUI for the program "mrtailor"
 * Copyright (C) 2012, Tim Gruene (tg@shelx.uni-ac.gwdg.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "MRTailorGui.h"
#include "PDBFileIn.h"
#include "PDBFileOut.h"
#include "Control.h"
#include "HeaderPart.h"
#include "QScoreInput.h"

#include "guiException.h"
#include "Exception.h"
#include "Sequence.h"
#include "QScores.h"
#include "ProSmart.h"

#include <QtGui>
#include <iostream>
#include <fstream>
#include <sstream>
#include <stdexcept>
#include <cerrno>
#include <qt4/QtGui/qboxlayout.h>

MRTailorGui::MRTailorGui(QWidget* parent)
: QWidget(parent), pdb_(NULL) {

    inputPDBfile_   = false;
    inputAlignment_ = false;
    
    aliLayout_  = new AlignLayout;
    runProsmart_ = new RunProsmart;

    QVBoxLayout* mainLayout = new QVBoxLayout;
    HeaderPart* headerpart  = new HeaderPart;
    PDBFileIn* pdbinput     = new PDBFileIn;
    QScoreInput* qscoreinput= new QScoreInput;
    PDBFileOut* pdboutput   = new PDBFileOut;
    Control* controls       = new Control;

    mainLayout->addWidget (headerpart);
    mainLayout->addWidget (pdbinput);
    mainLayout->addWidget (aliLayout_);
    mainLayout->addWidget (qscoreinput);
    mainLayout->addWidget (pdboutput);
    mainLayout->addWidget(runProsmart_);
    mainLayout->addWidget (controls);

    connect(aliLayout_, SIGNAL(new_sequence(const std::map<std::string, std::string>& )),
            this, SLOT(findSequence(const std::map<std::string, std::string>&)));
    connect(pdbinput, SIGNAL(new_file(QString)),
            this, SLOT(findSequence(QString)));
    connect (pdboutput, SIGNAL(new_file(QString)),
            this, SLOT(outputfile(QString)));
    connect (qscoreinput, SIGNAL(new_file(const QString&, int)), 
            this, SLOT(setQScoreData(const QString&, int)));
    
    connect(this, SIGNAL(sequenceMatch(std::string, std::string)),
            aliLayout_, SLOT(highlightMatches(std::string, std::string)));
    connect (controls, SIGNAL(runmrtailor()), this, SLOT (runmrtailor()));

    this->setLayout(mainLayout);
    setWindowTitle (tr ("Mr Tailor's GUI"));

}

MRTailorGui::~MRTailorGui() {
//    if (pdb_ != NULL)
//        delete pdb_;
//    pdb_ = NULL;
}

/*
 * Assume that among all sequences in the alignment there is a
 * longest one which contains one or more of the chains in the PDB-file
 */
void MRTailorGui::findSequence(const std::map<std::string, std::string>& seqs) {
    sequences_ = seqs;
    if (sequences_.empty()) {
        inputAlignment_ = false;
        // clear selection ...
        return;
    }
    inputAlignment_ = true;

    if (!inputPDBfile_ ) {
        return;
    }
    findSequence();
}

void MRTailorGui::findSequence(QString pdbfile) {
    inputPDBfile_ = false;
    try {
        pdb_ = new PDB(pdbfile.toStdString().c_str());
    }
    catch (guiException::PDBFormat& f) {
        inputPDBfile_ = false;
        return;
    }
    inputPDBfile_ = true;
    if (inputAlignment_ == true) {
        findSequence();
    }
}

void MRTailorGui::findSequence() {
    std::map<std::string, std::string>::const_iterator it;
    Sequence longestSeq;
    std::string chains;

    for(it = sequences_.begin(); it != sequences_.end(); ++it) {
        Sequence sequence (it->first, it->second);
        std::string testchains = pdb_->find_chains(sequence);
        if (!testchains.empty()) {
        }
        if (! testchains.empty() &&
                sequence.sequence_length() > longestSeq.sequence_length()) {
            longestSeq = sequence;
            chains     = testchains;
        }
        else {
            continue;
        }
    }
    if (chains.empty()) {
        QMessageBox* msgBox = new QMessageBox (QMessageBox::Critical,
                tr("No Sequence Match"), "", QMessageBox::Close);
        msgBox->setText(tr("None of the sequences in the alignment file"
                "are contained in any of the chains of the PDB file. "
                "Please change either PDB or alignment file."));
        msgBox->exec();
        return;
    }
    else {
        emit sequenceMatch(longestSeq.id(), chains);
    }
}




void MRTailorGui::runmrtailor() {
    /* sanity checks:
     * - target sequence must be set (from AlignLayout)
     * - model sequence must be found (in ids/chain from AlignLayout)
     * - model PDB must be set (from AlignLayout)
     * - target sequence and model sequence should differ (otherwise:
     *   issue a warning but continue).
     */
    Sequence modelseq  (aliLayout_->modelSequence());
    Sequence targetseq (aliLayout_->targetSequence());
    QScores qscores;
    std::string  pdbchain (aliLayout_->pdbChains());
    std::ofstream output;
    std::string::const_iterator it;
    
    modifiedChains_.clear();

    /* check name of output PDB file */
    if (output_file_.empty()) {
        QMessageBox* warningMsg = new QMessageBox(QMessageBox::Critical,
                tr("Missing Output File"), "",
                QMessageBox::Cancel);
        warningMsg->setText(tr("No output file specified. Hit 'Cancel' to provide output filename"
                " and run again."));
        warningMsg->exec();
        return;
    } else {
        output.open(output_file_.c_str());
        /* test that output file could be opened */
        if (!output.good()) {
            QMessageBox* warningMsg = new QMessageBox(QMessageBox::Critical,
                    tr("Unable to open file"), "", QMessageBox::Close);
            warningMsg->setText(tr("Could not open output file "));
            warningMsg->exec();
            return;
        }
    }
    
    /* both input PDB and alignment must be present */
    if (!inputPDBfile_ || ! inputAlignment_) {
        QMessageBox* warningMsg = new QMessageBox (QMessageBox::Information,
                tr("Warning"), tr("Missing Information"), QMessageBox::Close);
        if (!inputPDBfile_) {
            warningMsg->setText(tr("Please provide a Model PDB file"));
        }
        else {
            warningMsg->setText(tr("Please provide an alignment file"));
        }
        warningMsg->exec();
        return;
    }

    /* make sure target and model sequence should differ */
    if (aliLayout_->model_sequence_ID() == aliLayout_->target_sequence_ID()) {
        QMessageBox* msgBox = new QMessageBox  (QMessageBox::Warning,
                tr("Identical Sequcences"), "",
                QMessageBox::Yes|QMessageBox::No);
        msgBox->setText (tr ("The model sequence and target sequence are identical. "
        "Running MrTailor is not going to change anything in the output PDB file, "
        "except sorting the chain IDs alphabetically. Hit 'Yes' to continue,"
        "'No' to select a different target sequence ID."));
        int ret = msgBox->exec();
        if (ret == QMessageBox::No) {
            return;
        }
    }
    
    /* use qscore information if present */
    if (!qscoreFile_.isEmpty()) {
        try {
            qscores = QScores(qscoreFile_.toStdString().c_str(), targetseq);
        } catch (const Exception::QScoreFormat& e) {
            QMessageBox* msgBox = new QMessageBox(QMessageBox::Warning,
                    tr("Unable to read QScores file") + qscoreFile_,
                    "", QMessageBox::Ignore | QMessageBox::Retry);
            std::ostringstream msg;
            msg << "The file " << qscoreFile_.toStdString() << " could not"
                    << " be read as Clustal qscore-file.\n"
                    << "Please click \"Ignore\" to continue without"
                    << " qscore cut-off or \"Retry\" to enter a different file.";
            msgBox->setText(tr(msg.str().c_str()));
            int ret = msgBox->exec();
            if (ret == QMessageBox::Retry) {
                return;
            }
        } catch (const std::runtime_error& e) {
            QMessageBox* msgBox = new QMessageBox(QMessageBox::Warning,
                    tr("No Sequence match in ") + qscoreFile_,
                    "", QMessageBox::Ignore | QMessageBox::Retry);
            std::ostringstream msg;
            msg << "The sequence in the file \"" << qscoreFile_.toStdString()
                    << "\" does not match the selected Target Sequence.\n"
                    << "Please click \"Ignore\" to continue without"
                    << " qscore cut-off or \"Retry\" to enter a different file.";
            msgBox->setText(tr(msg.str().c_str()));
            int ret = msgBox->exec();
            if (ret == QMessageBox::Retry) {
                return;
            }
        }
    }

    /*
     * Checks passed, now run mrtailor
     * if 'proSmart' is checked: 
     * - run mrtailor
     * - run prosmart
     * - store output PDB?
     * - string fuer refmac aus prosmart Ergebnis
     */
    int longestChain(0);
    char prosmartReferenceChainID ('\0');
    for (it = pdbchain.begin(); it != pdbchain.end(); ++it) {
        Chain chain = pdb_->chain(*it);
        chain.mrtailor(modelseq, targetseq);

        if (chain.empty()) continue;
        if (chain.numResidues() > longestChain) {
            longestChain = chain.numResidues();
            prosmartReferenceChainID = *it;
        }

        if (qscores.initialised()) {
            chain.applyScores(qscores, qscoreCutoff_);
        }
                
        modifiedChains_.push_back(chain);
        modelseq.search_reset();
    }
    pdb_->writePDB(modifiedChains_, output);
    output.close();
 
    if (runProsmart_->runprosmart()) {
        callProsmart(prosmartReferenceChainID);
    }

    QMessageBox* msgBox = new QMessageBox (QMessageBox::Information,
                tr("Job's Done"), "", QMessageBox::Ok | QMessageBox::Close);
    std::ostringstream msg;
    msg <<  "The file " << output_file_ << " was written with "
            << modifiedChains_.size()
            << " modified chains to match the sequence "
            << aliLayout_->target_sequence_ID()
            << ". Hit 'Close' to exit the program, 'Ok' to continue.";
    msgBox->setText(tr(msg.str().c_str()));
    int ret = msgBox->exec();
    if (ret == QMessageBox::Close) {
        qApp->quit();
    }
    return;
}

void MRTailorGui::setQScoreData(const QString& filename, int cutoff) {
    qscoreFile_   = filename;
    qscoreCutoff_ = cutoff;
}

void MRTailorGui::outputfile(QString filename) {
    output_file_ = filename.toStdString();
}

int MRTailorGui::callProsmart(const char referenceChain) const {
    if (runProsmart_->pdbfile().isEmpty()) {
        QMessageBox* msgBox = new QMessageBox(QMessageBox::Warning,
                tr("No Refinement PDB file"),
                "", QMessageBox::Ignore | QMessageBox::Retry);
        std::ostringstream msg;
        msg << "Missing PDB file! In order to run prosmart you must provide a "
                "PDB file you are going to use for refinement.";
        msgBox->setText(tr(msg.str().c_str()));
        int ret = msgBox->exec();
        if (ret == QMessageBox::Retry) {
            return -1; /* needs catching */
        } else {
            runProsmart_->uncheck();
        }
    }

    const std::string refiPDB(runProsmart_->pdbfile().toStdString());
    const std::string outdir(runProsmart_->outdir().toStdString());
    ProSmart prosmart(refiPDB, pdb_->filename(), aliLayout_->targetSequence(), outdir);
    try {
        std::string chains = prosmart.chainsSearch();
        prosmart.run(chains, referenceChain);
    } catch (Exception::PDBSequence& e) {
        if (runProsmart_->forceSequence()) {
            std::string chains = prosmart.flexiSearchAndMatch(0.60);
            if (chains.empty()) {
                QMessageBox* msgBox = new QMessageBox(QMessageBox::Critical,
                    tr("No Chain Matching Sequence"), "", QMessageBox::Ok);
                std::ostringstream msg;
                msg << prosmart.warnNoSequences();
            msgBox->setText(tr(msg.str().c_str()));
            msgBox->exec();
            return -1;
            } else {
            prosmart.run(chains, referenceChain);
            }

        } else {
            QMessageBox* msgBox = new QMessageBox(QMessageBox::Critical,
                    tr("Sequence Mismatch"), "", QMessageBox::Ok);
            std::ostringstream msg;
            msg << "*** Error: No chain found in refinement PDB file "
                    << refiPDB << "\n"
                    << "           matching the target sequence. You can try the 'Force Sequence' button\n"
                    << "           for a more flexible search which will enforce the target \n"
                    << "           onto the refinement PDB file\n.";
            msgBox->setText(tr(msg.str().c_str()));
            msgBox->exec();
            return -1;
        }
    } catch (std::runtime_error& e) {
        QMessageBox* msgBox = new QMessageBox(QMessageBox::Critical,
                tr("Failure to Create Directory"), "", QMessageBox::Ok);
        std::ostringstream msg;
        msg << "---> Error: could not create directory " << e.what();
        msg << "     Error code: " << errno << std::endl;
        msgBox->setText(tr(msg.str().c_str()));
        msgBox->exec();
        return -1;
    } catch (std::logic_error& e) {
        QMessageBox* msgBox = new QMessageBox(QMessageBox::Critical,
                tr("Sequence Mismatch"), "", QMessageBox::Ok);
        std::ostringstream msg;
        msg << "The provided sequence \"" << target_.id() << "\" does not match any "
                "of the chains from the refinement PDB file. You can use e.g. pdbset "
                " to extract the sequences of all chains from the PDB file and compare "
                " to the sequence of \"" << target_.id() << "\" with clustalx or"
                "any other alignment program.\n";
        msgBox->setText(tr(msg.str().c_str()));
        msgBox->exec();
        return -1;
    }
    return 0;
}
