/* 
 * File:   ATOM.cpp
 * Author: tg
 * 
 * Created on October 28, 2011, 3:57 PM
 *
 * mrtailor: alignment based preparation of PDB files for MR.
 * Copyright (C) 2012, Tim Gruene (tg@shelx.uni-ac.gwdg.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "ATOM.h"
#include "Exception.h"
#include <sstream>
#include <iomanip>

ATOM::ATOM(const std::string atom_card)
: atom_card_(atom_card) {

    residue_type_ = atom_card_.substr(17,3);
    std::istringstream conv(atom_card_.substr(22,4));
    conv >> residue_number_;
}

void ATOM::renumber(unsigned int new_number) {

    // PDB only allows residue numbers < 9999
    if (new_number > 9999) {
        throw Exception::Bounds ("ATOM::renumber: residue number exceeds 9999");
    }
    std::ostringstream replacer;
    replacer << "" << std::setw(4) << new_number;
    atom_card_.replace(22, 4, replacer.str());
    
    residue_number_ = new_number;
}

/**
 * Return chain ID as stored in atom card
 * @return the chain ID of this atom
 */
char ATOM::chain() const {
    if (atom_card_.empty()) {
        return '-';
    }
    else {
        return atom_card_.at(21);
    }
}

/**
 * shift the residue number by @c delta (new = old+delta).
 * throws an exception, if new is negative or >9999
 * @param delta
 */
void ATOM::shift(int delta) {
    const int residue_field_with (4);
    const int residue_field_start(22);
    std::ostringstream resnum;
    
    residue_number_ += delta;

    resnum << std::setw(residue_field_with) << residue_number_;
    if (residue_number_ < 0 || residue_number_ > 9999) {
        residue_number_ -= delta;
        throw Exception::Bounds ("ATOM::shift - new residue has value "
                + resnum.str());
    }

    // replace residue number at position 22 with 4 chars
    atom_card_.replace(residue_field_start, residue_field_with, resnum.str()) = resnum.str();
}

/**
 * Replace residue name with @c newname. Only checks if newname consists of
 * exactly 3 characters, nothing else.
 * @param newname
 */
void ATOM::change_residue_name(const std::string& newname) {
    if (newname.size() != 3) {
        throw Exception::PDBFormat("ATOM::change_residue_name: residue must have three characters");
    }
    atom_card_.replace(17, 3, newname);
}

std::ostream& operator << (std::ostream& stream,
                            const ATOM& atom) {
    stream << atom.atom_card_ << '\n';
    return stream;
}
