/* 
 * File:   Chain.cpp
 * Author: tg
 * 
 * Created on October 31, 2011, 9:16 PM
 *
 * mrtailor: alignment based preparation of PDB files for MR.
 * Copyright (C) 2012, Tim Gruene (tg@shelx.uni-ac.gwdg.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Chain.h"
#include "QScores.h"
#include "Exception.h"
#include "Sequence.h"
#include "conversions.h"

#include <vector>

Chain::Chain()
: chainID_('\0') {
}

/**
 * Create a chain (list of Residue's) from atoms which are all supposed
 * to belong to one chain (increasing Residue number)
 * @param atoms
 */
Chain::Chain(const std::vector<ATOM>& atoms) {
    std::vector<ATOM>::const_iterator it;
    bool different_chains = false;

    std::list<ATOM> residue_atoms; // atoms within one reside
    int resnumber(0);
    for (it = atoms.begin(); it != atoms.end(); ++it) {
        different_chains |= (it->chain() != atoms.front().chain());
        if (it->residue_number() != resnumber) {
            if (!residue_atoms.empty()) {
                Residue new_residue(residue_atoms);
                chain_.push_back(new_residue);
            }
            resnumber = it->residue_number();
            residue_atoms.clear();
        }
        residue_atoms.push_back(*it);
    }

    // process last residue
    if (!residue_atoms.empty()) {
        Residue new_residue(residue_atoms);
        chain_.push_back(new_residue);
    }
    chainID_ = atoms.front().chain();

    if (different_chains == true) {
        throw Exception::Programming("Extraction of chain from PDB file");
    }
}

/**
 * Delete the residue with number @c residue_number,
 * i.e. replace it with an empty Residue (type '-')
 * @param residue_number
 */
void Chain::remove(unsigned int residue_number) {
    std::vector<Residue>::iterator it;
    for (it = chain_.begin(); it != chain_.end(); ++it) {
        if (it->residue_number() == residue_number) {
            *it = Residue(it->residue_number());
            break;
        }
    }
}

/**
 * change the Residue with number @c old_number to number @c new_number
 * @param old_number which Residue to chainge
 * @param new_number new residue number
 */
void Chain::renumber(unsigned int old_number, unsigned int new_number) {
    std::vector<Residue>::iterator it;

    for (it = chain_.begin(); it != chain_.end(); ++it) {
        int res_number = it->residue_number();
        if (res_number == old_number) {
            it->renumber(new_number);
            //! assume that residue number is unique
            return;
        }
    }
}

/**
 * checks if Residue numbers are sorted in increasing (or: not decreasing) order
 *
 * @return true, if residue are sorted by increasing residue numbers,
 * false otherwise.
 */
bool Chain::sorted() const {
    std::vector<Residue>::const_iterator it;

    for (it = chain_.begin() + 1; it != chain_.end(); ++it) {
        if (it->residue_number() < (it - 1)->residue_number())
            return false;
    }
    return true;
}

/**
 * modify this sequence according to sequence alignment
 * case 1: target_seq is empty: replace with empty residue
 * case 2: target_seq == model_seq: renumber residue to number target_seq
 * case 3: target_seq != model_seq: mutate to ALA and renumber
 * @param model_seq
 * @param target_seq
 */
void Chain::mrtailor(const Sequence& model_seq, const Sequence& target_seq) {
    std::vector<Residue>::iterator it;
    int residue_number;
    int sequence_index;
    char target_resname; // single letter residue name of target

    /* 
     * ensure sorted residue numbers, otherwise inc_search_resnum 
     * will cause trouble
     */
    if (!sorted()) {
        throw Exception::PDBFormat("Chain::mrtailor: unsorted chain");
    }

    for (it = chain_.begin(); it != chain_.end(); ++it) {
        residue_number = it->residue_number();
        sequence_index = model_seq.inc_search_resnum(residue_number);

        target_resname = target_seq.at(sequence_index);
        if (target_resname == '-') {
            // insert a gap into target
            *it = Residue(residue_number);
            continue;
        } else {
            if (target_resname != it->single_residue_letter()) {
                /* AA mismatch:
                 * convert residue name and strip coordinates to Alanine
                 */
                std::string newresname;
                newresname = Conversions::single2three(target_resname);
                it->alanine(newresname);
            }
            /* in both cases:
             * fix residue number
             */
            int target_resnum = target_seq.sequence_number(sequence_index);
            it->renumber(target_resnum);
        }
    }
}

/**
 * Make this chain's sequence match @c seq, assuming this is the correct string
 * with correct sequence numbers and gaps removed.
 * @param seq
 */
void Chain::force_seq(const std::string& seq) {
    std::vector<Residue>::iterator it;

    if (!sorted()) {
        throw Exception::PDBFormat("Chain::mrtailor: unsorted chain");
    }
    //! strip seq off gaps
    for (it = chain_.begin(); it != chain_.end(); ++it) {
        int resnumber = it->residue_number();
        char seq_aa = seq.at(resnumber-1); /* PDB file starts counting at '1' */
        
        if (it->single_residue_letter() == seq_aa) continue;

        std::string resname = Conversions::single2three(seq_aa);
        it->alanine(resname); 
    }
}

void Chain::applyScores(const QScores qscores, int cutOff) {
    std::vector<Residue>::iterator it;
    for (it = chain_.begin(); it != chain_.end(); ++it) {
        if (it->residue_number() > 0) {
            int n = it->residue_number();
            if (qscores.score(n - 1) < cutOff) {
                this->remove(n);
            }
        }
    }
}

/**
 * Comparison operator for Chains by ID
 * @param c1  first chain
 * @param c2  second chain
 * @return    id of c1 < id of c2
 */
bool operator<(const Chain& c1, const Chain& c2) {
    return c1.id() < c2.id();
}

/**
 * overloaded output operator
 * @param stream
 * @param chain
 * @return
 */
std::ostream& operator<<(std::ostream& stream, const Chain& chain) {
    std::vector<Residue>::const_iterator it;

    for (it = chain.chain_.begin(); it != chain.chain_.end(); ++it) {
        stream << *it;
    }
    return stream;
}
