/* 
 * File:   PDB.cpp
 * Author: tg
 * 
 * Created on October 30, 2011, 5:48 PM
 *
 * mrtailor: alignment based preparation of PDB files for MR.
 * Copyright (C) 2012, Tim Gruene (tg@shelx.uni-ac.gwdg.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "PDB.h"
#include "Sequence.h"
#include "ATOM.h"
#include "conversions.h"
#include "Exception.h"

#include <fstream>
#include <iostream>
#include <iomanip>
#include <vector>
#include <algorithm>
#include <map>
#include <string>
#include <sstream>

/**
 * initialise @c original_sequences from @c filename
 * and create sequences by calling @f sequence_extract
 * @param filename PDB file
 */
PDB::PDB(const char* filename): filename_ (filename) {
    std::ifstream inp (filename);
    if (! inp.is_open()) {
        std::cerr << "Could not open \"" << filename
                << "\"" << std::endl;
    }
    char* line;
    const unsigned short linewidth(100);

    line = new char [linewidth+1];

    while (true) {
        inp.getline(line, linewidth);
        if (!inp.good())    break;
        original_file_.push_back(line);
    }
    delete [] line;
    
    sequence_extract();
    
    if (chains_.empty() || sequences_.empty()) {
        throw Exception::PDBFormat("Unable to extract any coordinate data.");
    }
}

/**
 * read in sequence and chains from the PDB file
 */
void PDB::sequence_extract() {
    char chainID;
    char aa;
    int residue_number;
    ATOM atom;
    std::string sequence;
    std::string temp_chainID;
    std::vector<ATOM> chain_atoms;

    std::vector<std::string>::const_iterator it;

    chainID = '\0';
    residue_number = 0;
    
    for (it = original_file_.begin(); it != original_file_.end(); ++it) {
        if (it->substr(0, 4) != "ATOM") continue;
        atom = ATOM (*it);
        if (atom.chain() != chainID) {
            // new chain. Add current sequence to sequences, empty new sequence
            // and reset residue numberig (beginning at 0 again
            if (!sequence.empty()) {
                temp_chainID = chainID;
                Sequence new_sequence (temp_chainID, sequence);
                sequences_.push_back(new_sequence);
                sequence.clear();
                chains_.push_back(Chain(chain_atoms));
            }
            //  start a new chain at residue number 0
            residue_number = 0;
            chainID          = atom.chain();
            chain_atoms.clear();
        }
        chain_atoms.push_back(atom);

        // same chain, residue already processed -> next ATOM
        if (atom.residue_number() == residue_number) continue;
        // still within sequence, continue
        while (atom.residue_number() > residue_number+1) {
            sequence.append("-");
            ++residue_number;
        }
        aa = Conversions::three2single(atom.residue_type());
        sequence += aa;
        residue_number = atom.residue_number();

    }
    
    // the last chain will not have been processed
    if (!sequence.empty()) {
        temp_chainID = chainID;
        Sequence new_sequence(temp_chainID, sequence);
        sequences_.push_back(new_sequence);
        chains_.push_back(Chain(chain_atoms));

    }
}

/**
 * Find those chains in PDB file which match to @c seq
 * This codes relies on the residue numbers in the PDB file matching
 * the sequence numbering scheme
 * @param seq
 * @return a string containing the chain-IDs
 */
std::string PDB::find_chains(const Sequence& seq) const {
    std::string chains;
    std::vector<Sequence>::const_iterator it_seq;

    //! strip seq off gaps
    Sequence stripped_seq (seq);
    stripped_seq.strip_gaps();

    for (it_seq = sequences_.begin(); it_seq != sequences_.end(); ++it_seq) {
        //! if sequence in PDB-file is longer than search sequence: continue
        if (stripped_seq.contains(*it_seq)  ) {
            chains += it_seq->id();
        }
    }
    return chains;
}

/**
 * @c seq is assumed to be a coverage of PDB's chain and stripped off gaps
 * before comparison. Counts matches between this PDB's sequences and @c seq,
 * if relative number of matches is greater than @c perc_coverage
 * @param seq
 * @param perc_coverage
 * @return 
 */
std::string PDB::find_similar_chains(const Sequence& seq, double perc_coverage) {
    std::string chains("");

    std::vector<Sequence>::const_iterator it_seq;

    //! strip seq off gaps
    Sequence stripped_seq (seq);
    stripped_seq.strip_gaps();
    coverages_.clear();

    for (it_seq = sequences_.begin(); it_seq != sequences_.end(); ++it_seq) {
        //! if sequence in PDB-file is longer than search sequence: continue
        int matches, mismatches, gaps, insertions;
        count(stripped_seq, *it_seq, matches, mismatches, gaps, insertions);

        double coverage = 1.0*matches / (matches + mismatches);
        coverages_.push_back(coverage);
        
        if (coverage > perc_coverage) {
            chains += it_seq->id();
        }
    }
    return chains;
}

Chain PDB::chain(const char id) const {
    std::vector<Chain>::const_iterator it;

    for (it=chains_.begin(); it != chains_.end(); ++it) {
        if (it->id() == id) {
            return *it;
        }
    }
    return Chain();
}

void PDB::replace(const char id, const Chain& newchain) {
    std::vector<Chain>::iterator it;
    for (it = chains_.begin(); it != chains_.end(); ++it) {
        if (it->id() == id) {
            *it = newchain;
            break;
        }
    }
}

void PDB::writePDB(std::vector<Chain> newChains,
        std::ostream& outp) const {

    std::vector<std::string>::const_iterator it_header;
    std::vector<Chain>::const_iterator it_chain;
    std::string chainIDs;

    //! collect chain IDs in modified chains:
    for (it_chain=newChains.begin(); it_chain != newChains.end(); ++it_chain) {
        chainIDs += it_chain->id();
    }
    //! select chain from this which are not yet used.
    for (it_chain = chains_.begin(); it_chain != chains_.end(); ++it_chain) {
        if (chainIDs.find(it_chain->id()) == std::string::npos) {
            newChains.push_back(*it_chain);
        }
    }

    //! sort chains by ID
    std::sort(newChains.begin(), newChains.end());
    
    //! write header
    for (it_header = original_file_.begin(); it_header != original_file_.end(); ++it_header) {
        if (it_header->substr(0,4) == "ATOM") break;
        else {
            outp << *it_header << "\n";
        }
    }
    for (it_chain = newChains.begin(); it_chain != newChains.end(); ++it_chain) {
        outp << *it_chain << "TER\n";
    }

    //! write rest of file - continue at it_header position
    for (; it_header != original_file_.end(); ++it_header) {
        if (it_header->substr(0,4) == "ATOM" ||
                it_header->substr(0,3) == "TER" ||
                it_header->substr(0,6) == "ANISOU" ||
                it_header->substr(0,6) == "HETATM") continue;
        else {
            outp << *it_header << "\n";
        }
    }
}

std::string PDB::printFractions() const {
    if (chains_.size() != coverages_.size()) return "";
    std::ostringstream text;
    std::vector<Chain>::const_iterator it;
    for (it = chains_.begin(); it != chains_.end(); ++it) {
        unsigned int idx = it - chains_.begin();
        text << "   Chain \"" << it->id() << "\": " 
                << std::fixed << std::setprecision(1) << std::setw(5) 
                << 100.0*coverages_[idx] 
                << "%\n";
    }
    return text.str();
}