/* 
 * File:   ProSmart.cpp
 * Author: tg
 * 
 * Created on February 4, 2013, 5:00 PM
 */

#include <stdexcept>
#include <sys/stat.h>
#include <cstdlib>
#include <fstream>
#include <errno.h>
#include <unistd.h>
#include <sstream>

#include "ProSmart.h"
#include "Exception.h"

/**
 * 
 * @param refiPDB PDB file intended for refinment with refmac5
 * @param templatePDB PDB file to extract external restraints from
 * @param refiSequence Target sequence to find correct chain in refiPDB
 * @param outdir root for output directory of prosmart run
 */
ProSmart::ProSmart(const std::string& refiPDB,
        const std::string& templatePDB,
        const Sequence& refiSequence,
        const std::string outdir) :
        refiPDB_(refiPDB.c_str()), 
        refiPDBname_(refiPDB), 
        templatePDBname_(templatePDB),
        refiSequence_(refiSequence), 
        outdir_(outdir.empty()? "./prosmart" : outdir), 
        remove_temp_pdbfile_(false) {
}

ProSmart::~ProSmart() {
    if (remove_temp_pdbfile_) {
        unlink(refiPDBname_.c_str());
    }
}

/**
 * Search all chains in @c refiPDB_ exactly matching @c refiSequence_
 * @return 
 */
std::string ProSmart::chainsSearch() const {
    const std::string refiChains = refiPDB_.find_chains(refiSequence_);
    if (refiChains.empty()) {
        throw Exception::PDBSequence ("No Chains for ProSmart");
    }
    return refiChains;
}

/**
 * Running prosmart with options
 * #> prosmart -p1 targetPDB -c1 chainX -p2 referencePDB -c2 chainY
 * chainY is taken as first character in @c referenceChain it stems from the input sequence.
 * 
 * @param refiChains
 * @param referenceChain
 * @return 
 */
int ProSmart::run(const std::string& refiChains, const char templateChain) {
    int ret;
    
    prosmart_cmd_ = "prosmart";

    prosmart_cmd_ += " -p1 " + refiPDBname_ +
            " -p2 " + templatePDBname_;

    if (templateChain != '\0') {
        prosmart_cmd_ += " -c2 ";
        prosmart_cmd_ += templateChain;
    }

    if (refiChains.empty()) {
        throw Exception::PDBSequence("Running prosmart with empty target chain.");

    } else {
        if (!std::system(NULL)) {
            throw std::logic_error(prosmart_cmd_);
        }

        std::string::const_iterator itchains;
        for (itchains = refiChains.begin(); itchains != refiChains.end(); ++itchains) {
            std::string procmd(prosmart_cmd_);
            std::string dir(outdir_ + "_chain_" + *itchains);
            ret = mkdir(dir.c_str(), 493); /* = mode 755 */
            if (ret != 0) {
                if (errno != EEXIST) {
                    throw std::runtime_error(dir);
                }
            }
            procmd += " -c1 ";
            procmd += *itchains;
            procmd += " -o " + dir;
            ret = std::system(procmd.c_str());
            if (ret == -1) {
                throw Exception::SystemCall(procmd);
            }
        }
    }
    return ret;
}

/**
 * Find all chains in @c refiPDB_ which match @c refiSequence_ with a fraction
 * of at least @c coverage (0.0 <= coverage <=1.0)
 * @param coverage
 * @return 
 */
std::string ProSmart::flexiSearchAndMatch(const double coverage) {
    /* search PDB file for similar sequences */
    const std::string refiChains = refiPDB_.find_similar_chains(refiSequence_, coverage);
    Sequence stripped_seq (refiSequence_);
    stripped_seq.strip_gaps();
    std::string seq_string (stripped_seq.sequence());
    std::string::const_iterator it;
    std::vector<Chain> newchains;

    for (it = refiChains.begin(); it != refiChains.end(); ++it) {
        Chain chain = refiPDB_.chain(*it);
       
        chain.force_seq(seq_string);
        newchains.push_back(chain);
        refiPDB_.replace(chain.id(), chain);
    }
    
    /* write the modified chains to temporary file */
    std::ofstream out ("mrtailor_temporary_file.pdb");
    refiPDB_.writePDB(newchains, out);
    out.close();
    
    refiPDBname_ = "mrtailor_temporary_file.pdb";
    remove_temp_pdbfile_ = true;
        
    return refiChains;
}

/**
 * Print a warning message about the mismatch between sequence in the PDB-file
 * and the template sequence, including the fractions the template sequence
 * covers each chain.
 */
std::string ProSmart::warnNoSequences() const {
    std::ostringstream warning;
    warning << "ProSmart will not be executed because none of the sequences \n"
            << "   in the PDB file " << refiPDBname_ << " match the sequence \""
            << refiSequence_.id() << "\".\n\n"
            << "   Following is an estimate of the matches of \""
            << refiSequence_.id()
            << "\" vs. each chain in \"" << refiPDBname_ << "\":\n"
            << refiPDB_.printFractions()
            << "   If the PDB file intended for refinement stems from MR, you may want to check \n"
            << "   the mrtailor tutorial at \n"
            << "   http://shelx.uni-ac.gwdg.de/~tg/research/programs/mrtailor/tutorial\n"
            << std::endl;
    return warning.str();
}