/* 
 * File:   QScores.cpp
 * Author: tg
 * 
 * Created on February 14, 2012, 4:09 PM
 *
 * mrtailor: alignment based preparation of PDB files for MR.
 * Copyright (C) 2012, Tim Gruene (tg@shelx.uni-ac.gwdg.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "QScores.h"
#include "Exception.h"
#include "Chain.h"

#include <fstream>
#include <sstream>
#include <stdexcept>
#include <string>


/**
 * Read in a clustal qscore file (n columns with sequences and one last column 
 * with scores per residue. Store the sequence which is identical to 
 * @c target.
 * @param filename
 * @param target
 */
QScores::QScores(const char* filename, const Sequence& target) {
    int numSequences (0);
    const int maxLineLength (100);
    char line[maxLineLength];
    std::istringstream readline;
    std::string dummy;
    int score;
    
    std::vector<std::string> sequences;
    
    std::ifstream inp (filename);
    if (! inp.is_open()) {
        throw std::runtime_error ("Unable to open qscores file");
    }
    inp.getline(line, maxLineLength);
    
    readline.str(line);
    while (readline.good()) {
        readline >> dummy;
        ++numSequences;
    }
    --numSequences;
    sequences.resize(numSequences, "");
    
    while (inp.good()) {
        std::istringstream conv(line);
        for (int i = 0; i < numSequences; ++i) {
            conv >> dummy;
            sequences[i].push_back(dummy.at(0));
        }
        conv >> score;
        scores_.push_back(score);
        if (conv.fail()) {
            throw Exception::QScoreFormat("QSCore Format Error");
        }
        inp.getline(line, maxLineLength);
    }

    std::vector<std::string>::const_iterator it;
    for (it = sequences.begin(); it!= sequences.end(); ++it) {
        if (it->length() != scores_.size()) {
            scores_.clear();
            throw Exception::QScoreFormat("QScore Format Error"
                    ": Columns with different lengths");
        }
    }
    try {
        this->select(target, sequences);
        this->reduce();
    }    catch (std::runtime_error& e) {
        scores_.clear();
        throw;
    }
}

void QScores::select(const Sequence& sequence,
        const std::vector<std::string>& seqs) {
    std::vector<std::string>::const_iterator it;
    Sequence seq_copy (sequence);
    seq_copy.strip_gaps();
    for (it = seqs.begin(); it != seqs.end(); ++it) {
        Sequence cmp (std::string("qscore sequence"), *it);
        cmp.strip_gaps();
        if (cmp.contains(seq_copy)) {
            sequence_ = Sequence(std::string("qscore sequence"), *it);
            return;
        }
    }
    throw std::runtime_error ("QScores: no matching sequence in score-file");
}

/**
 * removes gaps from sequence and keeps @c scores_ in parallel
 */
void QScores::reduce() {
    if (sequence_.length() != scores_.size()) {
        throw Exception::Programming ("QScores: size mismatch");
    }
    std::string gap_free;
    std::vector<int> new_scores;
    for (unsigned int i = 0; i < sequence_.length(); ++i) {
        if (sequence_.at(i) != '-') {
            gap_free.push_back(sequence_.at(i));
            new_scores.push_back(scores_[i]);
        }
    }
    scores_ = new_scores;
    sequence_ = Sequence(sequence_.id(), gap_free);
    if (sequence_.length() != scores_.size()) {
        throw Exception::Programming ("QScores: size mismatch");
    }
}
