/* 
 * File:   Residue.cpp
 * Author: tg
 * 
 * Created on October 28, 2011, 6:14 PM
 *
 * mrtailor: alignment based preparation of PDB files for MR.
 * Copyright (C) 2012, Tim Gruene (tg@shelx.uni-ac.gwdg.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Residue.h"
#include "conversions.h"
#include "Exception.h"

/**
 * Initiate an empty residue, but with residue number
 * @param resnumber
 */
Residue::Residue(int resnumber)
: residue_number_ (resnumber), single_residue_letter_('-')
{}

Residue::Residue(const std::list<ATOM> atoms): atoms_(atoms) {
    std::string threelettercode = atoms_.front().residue_type();
    residue_number_ = atoms_.front().residue_number();
    single_residue_letter_ = Conversions::three2single(threelettercode);
}


/**
 * new residue number = old residue number + @c delta
 * @param delta difference new - old residue number
 */
void Residue::shift (int delta) {
    std::list<ATOM>::iterator it;
    residue_number_ += delta;

    for (it = atoms_.begin(); it != atoms_.end(); ++it) {
        it->shift(delta);
    }
}

/**
 *
 * @param new_number
 */
void Residue::renumber(unsigned int new_number) {

    std::list<ATOM>::iterator it;

    // only if there are any atoms, change ATOM cards
    if (single_residue_letter_ != '-') {
        for (it = atoms_.begin(); it != atoms_.end(); ++it) {
            it->renumber(new_number);
        }
    }
    // reset this counter, too
    residue_number_ = new_number;
}
/**
 * only leave N, CA, C, O, CB and change residue type to @c resname
 * (must be exactly 3 letters
 * @param resname New name for residue
 */
void Residue::alanine(const std::string& resname) {
    std::list<ATOM>::iterator it;
    if (resname.length() != 3) {
        throw Exception::Bounds ("Residue::alanine: resname must be 3 letters");
    }
    // make sure this is not an empty residue or GLY or ALA
//    if (single_residue_letter_ == '-' ||
//            single_residue_letter_ == 'G' ||
//            single_residue_letter_ == 'A') return;
    for (it = atoms_.begin(); it != atoms_.end();) {
        if (it->atom_name() == " CA " ||
                it->atom_name() == " N  " ||
                it->atom_name() == " C  " ||
                it->atom_name() == " O  " ||
                (resname != "GLY" && it->atom_name() == " CB ") ) {
            it->change_residue_name(resname);
            // must increase iterator
            ++it;
        } else {
            // erase atom and increase iterator
            it = atoms_.erase(it);
        }
    }
    single_residue_letter_ = Conversions::three2single(resname);
}

char Residue::single_residue_letter() const {
    return single_residue_letter_;
}

std::ostream & operator <<(std::ostream& stream,
            const Residue& residue) {
    std::list<ATOM>::const_iterator it;
    if (residue.single_residue_letter_ != '-') {
        for (it = residue.atoms_.begin(); it != residue.atoms_.end(); ++it) {
            stream << *it;
        }
    }
    return stream;
}
