/* 
 * File:   Sequence.cpp
 * Author: tg
 * 
 * Created on October 29, 2011, 11:35 AM
 *
 * mrtailor: alignment based preparation of PDB files for MR.
 * Copyright (C) 2012, Tim Gruene (tg@shelx.uni-ac.gwdg.de)
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include "Sequence.h"
#include "PDB.h"
#include "Exception.h"

#include <fstream>
#include <algorithm>

Sequence::Sequence() {
    it_search_index_ = sequence_numbers_.begin();
}

Sequence::~Sequence() {
}

Sequence::Sequence (const std::string& id, const std::string& seq)
: id_(id), sequence_(seq) {
    sequence_numbering();
    it_search_index_ = sequence_numbers_.begin();
}

Sequence::Sequence(const char* filename, const std::string& id): id_(id) {
    std::ifstream inp (filename);
    std::string word_read;

    if (!inp.is_open()) {
        std::cout << "Error opening sequence alignment file "
                << "\"" << filename << "\"\n";
        return;
    }
    int parts(0);

    while (inp.good()) {
        inp >> word_read;
        if (word_read == id_) {
            inp >> word_read;
            sequence_.append(word_read);
            ++parts;
        }
    }
    sequence_numbering();
    it_search_index_ = sequence_numbers_.begin();
    std::cout << "Read " << parts << " segments of " << id_ <<
            " with a total of " << this->length() << " aa" << std::endl;
}

/**
 * set up array of sequence numbers of which the index corresponds to @c
 * sequence_ and the corresponding entry corresponds to the real sequence
 * number (without gaps)
 */
void Sequence::sequence_numbering() {
   std::string::const_iterator it;
   sequence_numbers_ = std::vector<int> (sequence_.length(), 0);
   std::vector<int>::iterator it_tab (sequence_numbers_.begin());
   int idx;
   int residue_number(0);

   for (it = sequence_.begin(); it != sequence_.end(); ++it) {
       if (*it == '-')  continue;
       idx = it - sequence_.begin();
       //! set range from [tab to here) to residue number
       std::replace (it_tab, sequence_numbers_.begin()+idx, 0, residue_number);
       ++residue_number;
       it_tab = sequence_numbers_.begin()+idx;
   }
   idx = sequence_.length();
   std::replace (it_tab, sequence_numbers_.begin()+idx, 0, residue_number);
}

/**
 * remove all '-' from sequence
 */
void Sequence::strip_gaps() {
    std::string::iterator new_end;
    new_end = std::remove (sequence_.begin(), sequence_.end(), '-');
    sequence_ = std::string(sequence_.begin(), new_end);
}

unsigned int Sequence::gaps() const {
    unsigned int num (0);
    std::string::const_iterator it;

    for (it=sequence_.begin(); it != sequence_.end(); ++it) {
        if (*it == '-') {
            ++num;
        }
    }
    return num;
}

/**
 * check if @c other is contained in this. other may contain gaps,
 * but must not be longer than this
 * @param other
 * @return
 */
bool Sequence::contains(const Sequence& other) const {
    if (other.sequence_.length() > this->sequence_.length()) {
        return false;
    }

    std::string::const_iterator it;
    unsigned int idx;

    for (it = other.sequence_.begin(); it != other.sequence_.end(); ++it) {
        idx = it - other.sequence_.begin();
        if (*it == '-') continue;
        if (*it != sequence_.at(idx)) return false;
    }
    return true;
}

/**
 * Incremental search for residue number (i.e., stores the position
 * of the last search in order to continue there at a second call.
 * @param resnumber residue number to search for
 * @return index into sequence_ (i.e. taking gaps into account)
 */
int Sequence::inc_search_resnum (int resnumber) const {

    std::vector<int>::const_iterator it1, it2;
    it1 = it_search_index_;
    it2 = sequence_numbers_.end();

    it_search_index_ = std::find(it1, it2, resnumber);

    if (it_search_index_ == sequence_numbers_.end()) {
        it_search_index_ = sequence_numbers_.begin();
        throw Exception::Programming("Sequence::search_resnum: illegal residue number.");
    }
    return it_search_index_ - sequence_numbers_.begin();
}

/**
 * Output operator for Sequence, linewidth restricted to 80 chars.
 * @param stream
 * @param seq
 * @return stream with formatted sequence output
 */
std::ostream& operator<< (std::ostream& stream, const Sequence& seq) {
    int length (0);
    stream << "> " << seq.id_.substr(0, 78) << "\n";

    while (length < seq.sequence_.length()) {
        stream << seq.sequence_.substr(length, 80) << "\n";
        length += 80;
    }
    return stream;
}

/**
 * Assume both sequences are aligned (by gaps '-'), count how well they 
 * fit (no similarity weighting between amino acids).
 * @param other 
 * @param match identical a.a. 
 * @param misses mismatching a.a. present in both
 * @param gaps   @c other has a gap where @c this does not
 * @param inserts @c has a gap where @c other does not 
 */
void count(const Sequence& one, const Sequence& other, 
        int& match, int& misses, int& gaps, int& inserts) {
    match   = 0;
    misses  = 0;
    gaps    = 0;
    inserts = 0;
    
    int idx(0);
    int commonlength (std::min(one.length(), other.length()));

    for (idx = 0; idx < commonlength; ++idx) {
        if (one.at(idx) == '-' && other.at(idx) != '-') {
            ++inserts;
            continue;
        }
        if (one.at(idx) != '-' && other.at(idx) == '-') {
            ++gaps;
            continue;
        }
        // this should not happen, common gaps ignored
        if (one.at(idx) == '-' && other.at(idx) == '-') {
            continue;
        }
        if (one.at(idx) == other.at(idx)) {
            ++match;
            continue;
        }
        if (one.at(idx) != other.at(idx)) {
            ++misses;
            continue;
        }
    }
}